

# Optimization coordinates
function _opt(points_known_true::AbstractArray{<:Number,2},
	points_known_transf::AbstractArray{<:Number,2},
	points_to_transf::AbstractArray{<:Number,2};
	xyzguess=[0],opt_neigh=16)

	points_known_true = typeof(points_known_true)<:AbstractArray{Float64} ? points_known_true : convert(Array{Float64}, points_known_true)
	points_to_transf = typeof(points_to_transf)<:AbstractArray{Float64} ? points_to_transf : convert(Array{Float64}, points_to_transf)

    tree = BallTree(points_known_true)
    idxs, dists = knn(tree, points_to_transf, opt_neigh, true)

	out_coords = zeros(Float64,size(points_to_transf))
	Threads.@threads for i in 1:length(idxs)
		locs = view(points_known_transf,:,idxs[i])
        d = dists[i]
		initial_guess = points_known_transf[:,idxs[i][1]]
		if length(xyzguess)>1
			initial_guess = xyzguess[:,i]
		end

        opt = optimize(x->_mse_coords(x, locs, d), initial_guess)#, LBFGS())
        res = Optim.minimizer(opt)
        out_coords[:,i] .= res
	end
    out_coords
end



# Get surface normals
function _normals(ref_surf::AbstractArray{<:Number,2},nneigh::Number)

	ref_surf = typeof(ref_surf)<:AbstractArray{Float64} ? ref_surf : convert(Array{Float64}, ref_surf)
	tree = BallTree(ref_surf)
	idxs, dists = knn(tree, ref_surf, nneigh, true)

	ref_normals = zeros(Float64,size(ref_surf))

	last = nothing

	for i in 1:length(idxs)
		#pcapts = standardize(ZScoreTransform, ref_surf[:,idxs[i]], dims=2, scale=false)
		M = fit(PCA, view(ref_surf, :, idxs[i]), maxoutdim=3, pratio=1)
		ev = projection(M)

		if size(ev,2)>=3
			last = view(ev,:,3)
			ref_normals[:,i] .= last
		elseif size(ev,2)==2
			last = cross(view(ev,:,1),view(ev,:,2))
			ref_normals[:,i] .= last
		else # not best way to deal with 1 PC
			if last==nothing
				ref_normals[:,i] .= [0.,0.,1.]
			else
				ref_normals[:,i] .= last
			end
		end
	end

	# Iterate through all normals and check for consistency
	ok = false
	pre_loop = Array(1:length(idxs))
	to_loop = Int64[]
	visited = Int64[]

	xi = pre_loop[floor(Int,length(pre_loop)/2)]
	append!(visited,xi)
	to_loop = union(to_loop, idxs[xi])
	extra_loop = setdiff(to_loop,visited)
	while ok==false
		xi = extra_loop[1]
		append!(visited,xi)
		to_loop = union(to_loop, idxs[xi])
		extra_loop = setdiff(to_loop,visited)
		if length(extra_loop)==0
			ok=true
		end
	end

	for i in unique(to_loop)
		X = view(ref_normals,:,idxs[i])
		Y = fill(ref_normals[1,i],size(X))
		Y[2,:] .= ref_normals[2,i]
		Y[3,:] .= ref_normals[3,i]
		CD = colwise(CosineDist(), X, Y)
		ref_normals[:,idxs[i][CD .> 1]] .*= -1.0
	end

	ref_normals
end

# Error function
function _mse_coords(x::AbstractArray{Float64,1}, locations::AbstractArray{<:Number,2}, distances::AbstractArray{Float64,1})
    mse = 0.0
    for k in 1:length(distances)
        loc, d = [view(locations,:,k),distances[k]]
        distance_calculated = euclidean([x[1],x[2],x[3]],loc)
        mse += (distance_calculated - d)^2
    end
    mse / length(distances)
end

# Guess initial XYZ
function _xyzguess(coords_to_allocate::AbstractArray{<:Number,2}, ref_coords::AbstractArray{<:Number,2},
	 ref_transf_coords::AbstractArray{<:Number,2}, ref_surf_normals=nothing)

	ref_coords = typeof(ref_coords)<:AbstractArray{Float64} ? ref_coords : convert(Array{Float64}, ref_coords)
	coords_to_allocate = typeof(coords_to_allocate)<:AbstractArray{Float64} ? coords_to_allocate : convert(Array{Float64}, coords_to_allocate)
	tree = BallTree(ref_coords)
	idxs, dists = knn(tree, coords_to_allocate, 1, true)

	if ref_surf_normals==nothing
		filt = [idxs[x][1] for x in 1:length(idxs)]
		return ref_transf_coords[:,filt]
	else
		xyzguess = zeros(Float64,size(coords_to_allocate))

		for x in 1:length(idxs)
			filt = idxs[x][1]
			refs_coords = view(ref_coords,:,filt)
			refn_coords = view(ref_surf_normals,:,filt)
			ref_vect = view(coords_to_allocate,:,x)-refs_coords
			dotn_val = dot(refn_coords,ref_vect)
			if dotn_val<0
				xyzguess[3,x] = -1*dists[x][1]
			else
				xyzguess[3,x] = dists[x][1]
			end

			xyzguess[1,x] = ref_transf_coords[1,filt[1]]
			xyzguess[2,x] = ref_transf_coords[2,filt[1]]
		end

		return xyzguess
	end
end


"""
	unfold_error_ids(true_coords, transf_coords; nneigh=16, max_error=5)

Unfolding distorts the original distances between neighbor points. This
function give the IDs of the points above and below the `max_error` threshold.
Returns a tuple of two arrays. The first array with the ID of points that passed
the tests. The second array with the ID of points that failed during the tests.

## Parameters:

* `true_coords`   - coordinate matrix of shape (3,:) of the points before unfolding.
* `transf_coords` - coordinate matrix of shape (3,:) of the points after unfolding.
* `nneigh`        - number of nearest neighbors used to make the validations.
* `max_error`     - the maximum accepted absolute difference of the distances
  for the closest neighbors after unfolding.
"""
function unfold_error_ids(true_coords::AbstractArray{<:Number,2},
	 transf_coords::AbstractArray{<:Number,2};
	 nneigh=16, max_error=5)

	true_coords = typeof(true_coords)<:AbstractArray{Float64} ? true_coords : convert(Array{Float64}, true_coords)
    tree = BallTree(true_coords)
	idxs, dists = knn(tree, true_coords, nneigh, true)

	bad_ids = Int[]

	for x in 1:length(idxs)
		tdists = zeros(Float64,size(idxs[x]))
		for y in length(idxs[x])
			tdists[y] = euclidean(view(transf_coords,:,x),view(transf_coords,:,idxs[x][y]))
		end
		tdists .-= dists[x]
		tdists .= abs.(tdists)
		ids = findall(tdists .> max_error)

		if length(ids)>0
			push!(bad_ids,x)
			append!(bad_ids,idxs[x][ids])
		end
	end

	bad = unique(bad_ids)
	good = setdiff(Array(1:size(true_coords,2)),bad)
	good,bad
end

"""
	unfold_error_dists(true_coords, transf_coords; nneigh=16)

Unfolding distorts the original distances between neighbor points. This
function output the difference of the expected distance for each pair analyzed.
Can be used as input to boxplot to verify distortions.

## Parameters:

* `true_coords`   - coordinate matrix of shape (3,:) of the points before unfolding.
* `transf_coords` - coordinate matrix of shape (3,:) of the points after unfolding.
* `nneigh`        - number of nearest neighbors used to make the validations.
"""
function unfold_error_dists(true_coords::AbstractArray{<:Number,2},
	 transf_coords::AbstractArray{<:Number,2}; nneigh=16)

	true_coords = typeof(true_coords)<:AbstractArray{Float64} ? true_coords : convert(Array{Float64}, true_coords)
    tree = BallTree(true_coords)
	idxs, dists = knn(tree, true_coords, nneigh, true)

	error = Float64[]

	for x in 1:length(idxs)
		tdists = zeros(Float64,size(idxs[x]))
		for y in length(idxs[x])
			tdists[y] = euclidean(view(transf_coords,:,x),view(transf_coords,:,idxs[x][y]))
		end

		tdists .-= dists[x]
		tdists .= abs.(tdists)

		append!(error,tdists[2:end])
	end

	error
end
